#include "cmdline.hh"
#include <cstdlib>
#include <error.h>

// These pointers have been globally declared by argp.h.
const char *argp_program_version {PROGNAME " " PROGVER};
const char *argp_program_bug_address {PROGEMAIL};

namespace {

    struct Cmdline_args_manager {
        bool has_set_name_of_number;
        Driver::Cmdline_args *ptr_cmdline_args;
        Cmdline_args_manager(Driver::Cmdline_args *);
    };

    Cmdline_args_manager::Cmdline_args_manager(Driver::Cmdline_args *const p):
      has_set_name_of_number {false}, ptr_cmdline_args {p}
    {}

    void cmdline_if_error_then_report_and_exit(const error_t code)
    { if (code) error(EXIT_FAILURE, code, "error"); }

    error_t cmdline_parse
      (const int key, char *const arg, argp_state *const state)
    {
        Cmdline_args_manager &manager
          {*static_cast<Cmdline_args_manager *>(state->input)};
        Driver::Cmdline_args &cmdline_args {*manager.ptr_cmdline_args};
        switch(key) {
        case 'l':
            cmdline_args.facets_to_print.push_back
              (Driver::Facet_to_print::lo);
            break;
        case 'h':
            cmdline_args.facets_to_print.push_back
              (Driver::Facet_to_print::hi);
            break;
        case 'd':
            cmdline_args.facets_to_print.push_back
              (Driver::Facet_to_print::delta);
            break;
        case 'm':
            cmdline_args.facets_to_print.push_back
              (Driver::Facet_to_print::mid);
            break;
        case 'e':
            cmdline_args.facets_to_print.push_back
              (Driver::Facet_to_print::dev);
            break;
        case 'p':
            cmdline_args.facets_to_print.push_back
              (Driver::Facet_to_print::prec);
            break;
        case 'o':
            cmdline_args.facets_to_print.push_back
              (Driver::Facet_to_print::order);
            break;
        case 'O':
            cmdline_args.order_max = std::strtol(arg, 0, 0);
            break;
        case 'n':
            cmdline_args.prefer_negative_branch = true;
            break;
        case 'q':
            cmdline_args.verbosity = -1;
            break;
        case 'v':
            if (cmdline_args.verbosity >= 0) ++cmdline_args.verbosity;
            break;
        case ARGP_KEY_ARG:
            switch (state->arg_num) {
            case 0:
                {
                    const long precision0 {std::strtol(arg, 0, 0)};
                    if (precision0 < 0)
                      cmdline_if_error_then_report_and_exit(EINVAL);
                    cmdline_args.precision = precision0;
                }
                break;
            case 1:
                cmdline_args.name_of_number = arg;
                manager.has_set_name_of_number = true;
                break;
            default:
                cmdline_args.parameters.push_back(std::strtol(arg, 0, 0));
            }
            break;
        case ARGP_KEY_END:
            if (!manager.has_set_name_of_number) argp_usage(state);
            break;
        default:
            return ARGP_ERR_UNKNOWN;
        }
        return 0;
    }

}

Driver::Cmdline_args::Cmdline_args():
  prefer_negative_branch {false}, order_max {-1}, verbosity {0}
{}

Driver::Cmdline_args Driver::read_cmdline
  (const int argc, char **const argv)
{
    Cmdline_args result;
    Cmdline_args_manager manager(&result);
    const argp parser {
        cmdline_options, cmdline_parse, cmdline_args_doc, cmdline_doc, 0, 0, 0
    };
    const error_t error_code {argp_parse(&parser, argc, argv, 0, 0, &manager)};
    cmdline_if_error_then_report_and_exit(error_code);
    if(result.facets_to_print.empty() && result.verbosity >= 0) {
        error(
            0, 0,
            "warning: "
            "no printing options have been specified, "
            "so number will be printed"
        );
    }
    return result;
}

