#ifndef CMDLINE_HH
#define CMDLINE_HH

#include "cdef.hh"
#include <list>
#include <string>
#include <argp.h>
#include <gmpxx.h>

namespace Driver {

    constexpr char cmdline_doc[] {
        PROGNAME
        " -- calculate a (typically irrational) real number, "
        "binding it within arbitrarily tight bounds, "
        "printing it in hexadecimal"
        "\v"
        "NUMBER-OF-BITS-OF-PRECISION is the minimum number of bits "
        "to use to calculate and store the result's fractional part.  "
        "It is not restricted by the number of bits the CPU's architecture "
        "affords, but may be arbitrarily large.\n"
        "\n"
        "NAME-OF-CONSTANT [ARGS ...] must be one of these:\n"
        "\n"
        "  quadroot N S P Q        for (P + S*sqrt(N))/Q\n"
        "  power    N M P Q        for (P/Q)^(N/M)\n"
        "  twopi                   for twice Archimedes' constant\n"
        "  sqrttwopi               for the square root of twopi\n"
        "  log      P Q            for the natural logarithm of P/Q\n"
        "  logtwopi                for the natural logarithm of twopi\n"
        "\n"
        "In normal usage,\n"
        "\n"
        "  * at least one of [-delhm] will be specified and\n"
        "  * a [--] will be given to terminate options processing "
        "(for otherwise, the program will misinterpret "
        "negative parameters as options).\n"
        "\n"
        "Example, calculating -sqrt(2) to 128 bits:\n"
        "            " PROGNAME " -lh -- 0x80 quadroot 2 -1 0 1"
    };
    constexpr char cmdline_args_doc[] {
        "NUMBER-OF-BITS-OF-PRECISION NAME-OF-CONSTANT [ARGS ...]"
    };
    constexpr argp_option cmdline_options[] {
        {
            "lo", 'l', 0, 0,
            "Print the lower bound", 0
        },
        {
            "hi", 'h', 0, 0,
            "Print the upper bound", 0
        },
        {
            "delta", 'd', 0, 0,
            "Print the difference between upper and lower bounds", 0
        },
        {
            "mid", 'm', 0, 0,
            "Print the midpoint between lower and upper bounds", 0
        },
        {
            "dev", 'e', 0, 0,
            "Print the maximum deviation from the midpoint", 0
        },
        {
            "check-precision", 'p', 0, 0,
            "Print the precision with which calculation "
            "has actually been done", 0
        },
        {
            "order", 'o', 0, 0,
            "Print the order to which the mathematical series (if any) "
            "has been computed", 0
        },
        {
            "max-order", 'O', "MAX-ORDER", 0,
            "To the extent possible, calculate no farther than MAX-ORDER", 0
        },
        {
            "prefer-negative-branch", 'n', 0, 0,
            "Where relevant, "
            "prefer the mathematical function's negative branch", 0
        },
        {
            "quiet", 'q', 0, 0,
            "Report neither warnings nor diagnostics", 0
        },
        {
            "verbose", 'v', 0, 0,
            "Report extra warnings and diagnostics "
            "(reserved for future implementation, "
            "this option is safe to use but presently lacks effect)", 0
        },
        {0, 0, 0, 0, 0, 0}
    };

    enum class Facet_to_print;
    struct Cmdline_args;
    // The header argp.h does not allow const char *const *argv.
    Cmdline_args read_cmdline(int argc, char **argv);

}

enum class Driver::Facet_to_print {
    none, // unused
    lo, hi, delta, mid, dev,
    prec, order // pseudofacets
};

struct Driver::Cmdline_args {
    mp_bitcnt_t precision;
    std::string name_of_number;
    std::list<long> parameters;
    bool prefer_negative_branch;
    long order_max;
    std::list<Facet_to_print> facets_to_print;
    int verbosity;
    Cmdline_args();
};

#endif

