#include "divrem.hh"
#include <cstdlib>

//                            Interval::
//                  CPU       divrem()    Adjustment
//
// (+14)/(+3) == +4 rem +2 == +4 rem +2    0:0
// (+14)/(-3) == -4 rem +2 == -5 rem -1   -1:+q
// (-14)/(+3) == -4 rem -2 == -5 rem +1   -1:+q
// (-14)/(-3) == +4 rem -2 == +4 rem -2    0:0
// (+18)/(+3) == +6 rem  0 == +6 rem  0    0:0
// (+18)/(-3) == -6 rem  0 == -6 rem  0    0:0
// (-18)/(+3) == -6 rem  0 == -6 rem  0    0:0
// (-18)/(-3) == +6 rem  0 == +6 rem  0    0:0
// (  0)/(+3) ==  0 rem  0 ==  0 rem  0    0:0
// (  0)/(-3) ==  0 rem  0 ==  0 rem  0    0:0
// (  0)/(+3) ==  0 rem  0 ==  0 rem  0    0:0
// (  0)/(-3) ==  0 rem  0 ==  0 rem  0    0:0
//
// that P/Q == A rem R implies that P == A*Q + R
//
//

Interval::Quotrem<long, long> Interval::divrem
  (const long p, const long q)
{
    Quotrem<long, long> result;
    if (!q) throw Division_by_zero {};
    const std::ldiv_t a {std::ldiv(p, q)};
    result.quot = a.quot;
    result.rem  = a.rem;
    if (result.rem && ((p < 0 && q >= 0) || (p >= 0 && q < 0)))
      result.quot -= 1, result.rem  += q;
    return result;
}

Interval::Quotrem<mpz_class, long> Interval::divrem
  (const mpz_class &p, const long q)
{
    Quotrem<mpz_class, long> result;
    if (!q) throw Division_by_zero {};
    mpz_class a_rem;
    mpz_fdiv_qr_ui(
        result.quot.get_mpz_t(), a_rem.get_mpz_t(),
        p.get_mpz_t(), std::abs(q)
    );
    result.rem = a_rem.get_si();
    // (+14)/(-3), computed as (+14)/(+3) == +4 rem +2, should be -5 rem -1
    // (-14)/(-3), computed as (-14)/(+3) == -5 rem +1, should be +4 rem -2
    // (+ 2)/(-3), computed as (+ 2)/(+3) ==  0 rem +2, should be -1 rem -1
    // (- 2)/(-3), computed as (- 2)/(+3) == -1 rem +1, should be  0 rem -2
    if (q < 0) {
        result.quot = -result.quot;
        if (result.rem) result.quot -= 1, result.rem  += q;
    }
    return result;
}

Interval::Quotrem<mpz_class, mpz_class> Interval::divrem
  (const mpz_class &p, const mpz_class &q)
{
    Quotrem<mpz_class, mpz_class> result;
    if (!q) throw Division_by_zero {};
    mpz_fdiv_qr(
        result.quot.get_mpz_t(), result.rem.get_mpz_t(),
        p.get_mpz_t(), q.get_mpz_t()
    );
    return result;
}

