#include "log.hh"
#include "gcd.hh"
#include "twopi.hh"
#include <utility>

// To calculate a logarithm in as few iterations as practicable is not this
// module's purpose, but rather to calculate a logarithm according to a
// specific, consistently applied algorithm.

namespace {
    // For 0 <= x < 1, calculate -log(1-x).  The parameter geometric
    // must equal 1/(1-x) = 1 + x + x^2 + x^3 + ... as nearly as the
    // caller can manage, but the method by which the caller achieves
    // best accuracy in this is for the caller to choose and implement.
    Interval::Lohi calc_log_impl(
        const mp_bitcnt_t precision,
        const Interval::Lohi x, const Interval::Lohi geometric,
        Interval::Order *const ptr_order
    )
    {
        Interval::Prelohi result {0};
        const mpz_class one {mpz_class {1} << precision};
        if (x.lo() < 0 || x.hi() >= one || geometric.lo() < one)
          throw Interval::Domain_straddle {};
        Interval::Lohi preaddend {x};
        Interval::Lohi quarry {0};
        const long order_overmax
          { ptr_order->max() <  0 ? -1 : ptr_order->max() + 1 };
        for (ptr_order->reset().increment(1);; ptr_order->increment(1)) {
            // This loop works in a nonobvious way.  The addend is
            // calculated at the start of this iteration for use during
            // the next iteration, more than a full cycle later.  The
            // reason for such peculiar sequencing is that the addend
            // is incidentally useful a cycle early, during calculation
            // of the geometric error bound.  A side effect of the
            // peculiar sequencing is that the loop's initial iteration
            // serves mainly to initialize variables; the initial
            // iteration does compute a proper error bound but is not
            // otherwise a proper part of the series calculation.
            // (When the writer reads slightly clever code like this
            // written by someone else, he wants to know what has
            // inspired that someone to the cleverness.  However, the
            // cleverness -- such as it is -- in the present code did
            // not occur to the writer in any inscrutable burst of
            // inspiration.  Rather, the writer wrote the code first in
            // a clumsy, nonclever way, got the code working, then
            // iteratively refactored, and then tidied up.  That the
            // writer has years of practical experience in refactoring
            // helped, as did the writer's general coding practice --
            // which the writer suggests more C++ programmers should
            // emulate -- of using the break statement to rotate loop
            // bodies to find neater points of entry and exit; but the
            // writer is as suprised as anyone else by the slight
            // cleverness of the end result.)
            const Interval::Lohi addend {preaddend / ptr_order->count()};
            const Interval::Lohi error_bound
              { quarry + ((addend * geometric) >> precision) };
            if (quarry.lo() <= result.lo() && ptr_order->count() > 1)
              { ptr_order->increment(-2); break; }
            result.lo() = quarry.lo();
            if (error_bound.hi() < result.hi() || ptr_order->count() <= 1)
              result.hi() = error_bound.hi();
            if (
                order_overmax >= 0 &&
                ptr_order->count() >= order_overmax
            ) { ptr_order->increment(-1); break; }
            // Here is where the addend enters the main calculation, as
            // a constituent of the quarry.  Notice that the quarry is
            // not used until after the addend's storage has been
            // overwritten, after which the addend's value survives
            // only as a constituent of the quarry.
            quarry += addend;
            preaddend *= x;
            preaddend >>= precision;
        }
        return Interval::Lohi {result};
    }
}

Interval::Lohi Interval::calc_log(
    const mp_bitcnt_t precision,
    long p, long q,
    Order *const ptr_order
)
{
    reduce(&p, &q);
    if (p <= 0) throw Forbidden_negative {};
    const bool do_negate {p <= q};
    if (!do_negate) std::swap(p, q);
    const Lohi x {Lohi {mpz_class {q - p} << precision} / q};
    const Lohi geometric {
        // 1/(1-x) = 1/(p/q) = q/p
        Lohi {mpz_class {q} << precision} / p
    };
    const Lohi result
      { calc_log_impl(precision, x, geometric, ptr_order) };
    return do_negate ? -result : result;
}

Interval::Lohi Interval::calc_logtwopi
  (const mp_bitcnt_t precision, Order *const ptr_order)
{
    const Lohi twopi {
        [&precision]() -> Lohi {
            Order order_twopi(-1);
            return calc_twopi(precision, &order_twopi);
        }()
    };
    return calc_log_impl(
        precision,
        // x = 1 - 1/twopi; geometric = 1/(1-x) = 1/(1/twopi) = twopi
        (mpz_class {1} << precision) -
          (mpz_class {1} << (precision << 1)) / twopi,
        twopi,
        ptr_order
    );
}

