#include "lohi.hh"
#include "divrem.hh"
#include <utility>
#include <algorithm>

namespace {

    inline const Interval::Lohi &check(const Interval::Lohi &a)
    { if (a.hi() < a.lo()) throw Interval::Lohi_bad_state {}; return a; }

    inline Interval::Lohi &check(Interval::Lohi &a)
    { if (a.hi() < a.lo()) throw Interval::Lohi_bad_state {}; return a; }

    template<class T> inline void impl_multiply
      (mpz_class *const ptr_lo, mpz_class *const ptr_hi, const T &multiplicand)
      noexcept
    {
        if (multiplicand < 0) std::swap(*ptr_lo, *ptr_hi);
        *ptr_lo *= multiplicand, *ptr_hi *= multiplicand;
    }

    template<class T> inline void impl_divide
      (mpz_class *const ptr_lo, mpz_class *const ptr_hi, const T &divisor)
    {
        if (divisor < 0) std::swap(*ptr_lo, *ptr_hi);
        *ptr_lo = Interval::divrem(*ptr_lo, divisor).quot;
        const Interval::Quotrem<mpz_class, T>
          quotrem_hi {Interval::divrem(*ptr_hi, divisor)};
        *ptr_hi = quotrem_hi.quot + bool(quotrem_hi.rem);
    }

}

Interval::Prelohi::Prelohi() noexcept:
  lo_ {0}, hi_ {0}
{}

Interval::Prelohi::Prelohi(const mpz_class &n1) noexcept:
  lo_ {n1}, hi_ {n1}
{}

Interval::Prelohi::Prelohi
  (const mpz_class &lo1, const mpz_class &hi1) noexcept:
  lo_ {lo1}, hi_ {hi1}
{}

Interval::Prelohi &Interval::Prelohi::swap() noexcept
{ std::swap(lo_, hi_); return *this; }

Interval::Lohi::Lohi() noexcept:
  lo_ {0}, hi_ {0}
{}

Interval::Lohi::Lohi(const mpz_class &n1) noexcept:
  lo_ {n1}, hi_ {n1}
{}

Interval::Lohi::Lohi(const mpz_class &lo1, const mpz_class &hi1):
  lo_ {lo1}, hi_ {hi1}
{ check(*this); }

Interval::Lohi::Lohi(const Prelohi &prelohi1):
  lo_ {prelohi1.lo()}, hi_ {prelohi1.hi()}
{ check(*this); }

Interval::Lohi &Interval::Lohi::operator=(const mpz_class &b)
{ lo_ = b, hi_ = b; return check(*this); }
// It looks as though this function had memory trouble, but closer
// inspection of each degenerate case discovers that none of the
// degenerates -- not even &b == &lo_ -- actually manifests
// the trouble.

Interval::Lohi &Interval::Lohi::operator+=(const Lohi &b)
{ lo_ += b.lo(), hi_ += b.hi(); return check(*this); }

Interval::Lohi &Interval::Lohi::operator+=(const mpz_class &b)
{
    if (&b == &lo_) hi_ += b, lo_ += b;
    else            lo_ += b, hi_ += b;
    return check(*this);
}

Interval::Lohi &Interval::Lohi::operator-=(const Lohi &b_)
{
    const Lohi &b {&b_ == this ? Lohi {b_} : b_};
    lo_ -= b.hi(), hi_ -= b.lo(); return check(*this);
}

Interval::Lohi &Interval::Lohi::operator-=(const mpz_class &b)
{
    if (&b == &lo_) hi_ -= b, lo_ -= b;
    else            lo_ -= b, hi_ -= b;
    return check(*this);
}

Interval::Lohi &Interval::Lohi::operator*=(const Lohi &b_)
{
    const Lohi &b {&b_ == this ? Lohi {b_} : b_};
    const bool is_this_negative {hi_ < 0};
    const bool does_this_straddle {!is_this_negative && lo_ < 0};
    const bool is_b_negative {b.hi() < 0};
    const bool does_b_straddle {!is_b_negative && b.lo() < 0};
    if (is_b_negative) std::swap(lo_, hi_);
    if (does_this_straddle) {
        if (does_b_straddle) {
            const mpz_class lo0 {lo_};
            lo_ = std::min(lo0 * b.hi(), hi_ * b.lo());
            hi_ = std::max(lo0 * b.lo(), hi_ * b.hi());
        }
        else {
            if (is_b_negative) lo_ *= b.lo(), hi_ *= b.lo();
            else               lo_ *= b.hi(), hi_ *= b.hi();
        }
    }
    else {
        if (is_this_negative) {
            if(does_b_straddle) hi_ = lo_;
            lo_ *= b.hi(), hi_ *= b.lo();
        }
        else {
            if(does_b_straddle) lo_ = hi_;
            lo_ *= b.lo(), hi_ *= b.hi();
        }
    }
    return check(*this);
}

Interval::Lohi &Interval::Lohi::operator*=(const mpz_class &b_)
{
    const mpz_class &b {&b_ == &lo_ || &b_ == &hi_ ? mpz_class {b_} : b_};
    impl_multiply(&lo_, &hi_, b); return check(*this);
}

Interval::Lohi &Interval::Lohi::operator*=(const long b)
{ impl_multiply(&lo_, &hi_, b); return check(*this); }

Interval::Lohi &Interval::Lohi::operator/=(const mpz_class &b_)
{
    const mpz_class &b {&b_ == &lo_ || &b_ == &hi_ ? mpz_class {b_} : b_};
    impl_divide(&lo_, &hi_, b); return check(*this);
}

Interval::Lohi &Interval::Lohi::operator/=(const long b)
{ impl_divide(&lo_, &hi_, b); return check(*this); }

Interval::Lohi &Interval::Lohi::operator<<=(const mp_bitcnt_t c)
{
    lo_ <<= c, hi_ <<= c;
    return check(*this);
}

Interval::Lohi &Interval::Lohi::operator>>=(const mp_bitcnt_t c)
{
    const bool hi_adjustment {!mpz_divisible_2exp_p(hi_.get_mpz_t(), c)};
    lo_ >>= c, hi_ >>= c;
    hi_ += hi_adjustment;
    return check(*this);
}

Interval::Lohi &Interval::Lohi::square()
{
    const bool is_this_negative {hi_ < 0};
    const bool does_this_straddle {!is_this_negative && lo_ < 0};
    if (is_this_negative) std::swap(lo_, hi_);
    if (does_this_straddle) {
        hi_ = hi_ < -lo_ ? lo_ * lo_ : hi_ * hi_;
        lo_ = 0;
    }
    else lo_ *= lo_, hi_ *= hi_;
    return check(*this);
}

Interval::Lohi &Interval::Lohi::negate()
{ std::swap(lo_, hi_); lo_ = -lo_, hi_ = -hi_; return check(*this); }

Interval::Lohi &Interval::Lohi::invert_against(const mpz_class &b_)
{
    const mpz_class &b {&b_ == &lo_ || &b_ == &hi_ ? mpz_class {b_} : b_};
    if (lo_ < 0 && hi_ >= 0) throw Interval::Division_by_zero {};
    if (b >= 0) std::swap(lo_, hi_);
    lo_ = Interval::divrem(b, lo_).quot;
    const Interval::Quotrem<mpz_class, mpz_class>
      quotrem_hi {Interval::divrem(b, hi_)};
    hi_ = quotrem_hi.quot + bool(quotrem_hi.rem);
    return check(*this);
}

Interval::Lohi Interval::operator+(const Lohi &a)
{ return check(a); }

Interval::Lohi Interval::operator-(Lohi a)
{ return a.negate(); }

Interval::Lohi Interval::operator+(Lohi a, const Lohi &b)
{ return a += b; }

Interval::Lohi Interval::operator+(Lohi a, const mpz_class &b)
{ return a += b; }

Interval::Lohi Interval::operator+(const mpz_class &b, Lohi a)
{ return a += b; }

Interval::Lohi Interval::operator-(Lohi a, const Lohi &b)
{ return a -= b; }

Interval::Lohi Interval::operator-(Lohi a, const mpz_class &b)
{ return a -= b; }

Interval::Lohi Interval::operator-(const mpz_class &b, Lohi a)
{ return a.negate() += b; }

Interval::Lohi Interval::operator*(Lohi a, const Lohi &b)
{ return a *= b; }

Interval::Lohi Interval::operator*(Lohi a, const mpz_class &b)
{ return a *= b; }

Interval::Lohi Interval::operator*(Lohi a, const long b)
{ return a *= b; }

Interval::Lohi Interval::operator*(const mpz_class &b, Lohi a)
{ return a *= b; }

Interval::Lohi Interval::operator*(const long b, Lohi a)
{ return a *= b; }

Interval::Lohi Interval::operator/(Lohi a, const mpz_class &b)
{ return a /= b; }

Interval::Lohi Interval::operator/(Lohi a, const long b)
{ return a /= b; }

Interval::Lohi Interval::operator/(const mpz_class &b, Lohi a)
{ return a.invert_against(b); }

Interval::Lohi Interval::operator<<(Lohi a, const mp_bitcnt_t c)
{ return a <<= c; }

Interval::Lohi Interval::operator>>(Lohi a, const mp_bitcnt_t c)
{ return a >>= c; }

Interval::Lohi Interval::square(Lohi a)
{ return a.square(); }

mpz_class Interval::square(mpz_class a)
{ return a *= a; }

long Interval::square(long a)
{ return a *= a; }

Interval::Lohi &Interval::square(Lohi *const p)
{ return p->square(); }

mpz_class &Interval::square(mpz_class *const p)
{ return *p *= *p; }

long &Interval::square(long *const p)
{ return *p *= *p; }

Interval::Lohi &Interval::negate(Lohi *const p)
{ return p->negate(); }

mpz_class &Interval::negate(mpz_class *const p)
{ return *p = -*p; }

long &Interval::negate(long *const p)
{ return *p = -*p; }

mpz_class Interval::delta(const Lohi &a)
{ return a.hi() - a.lo(); }

mpz_class Interval::mid(const Lohi &a)
{
    mpz_class sum {a.hi() + a.lo()};
    const bool is_sum_negative = sum < 0;
    if (is_sum_negative) sum = -sum;
    const mpz_class half {(sum >> 1) + ((sum.get_ui() & 3) == 3)};
    return is_sum_negative ? -half : half;
}

mpz_class Interval::dev(const Lohi &a)
{
    const mpz_class mid1 = mid(a);
    return std::max(mid1 - a.lo(), a.hi() - mid1);
}

