#ifndef LOHI_HH
#define LOHI_HH
#include "exceptions.hh"
#include <gmpxx.h>

// Interval::Lohi is a type to represent an integer of arbitrary size
// when the integer's exact value is unknown but lower and upper bounds,
// inclusive, between which the integer must lie are known.
//
// Summary of operations:
//
//    +lohi       -lohi
//     lohi + m    lohi - m    m - lohi
//     lohi * n    lohi / n
//     lohi << k   lohi >> k
//     square(n)
//     square(*n)  negate(*n)
//     += -= *= /= <<= >>=
//
// The m may be of Interval::Lohi or mpz_class type.
// The n may be of Interval::Lohi, mpz_class or long type.
// The k is of GMP's mp_bitcnt_t type.
//
// The square(lohi1) function is normally preferable to lohi1*lohi1
// because the latter -- which does not realize that the two lohis are
// the same -- can return a negative lower bound.
//
//

namespace Interval {
    struct Lohi_bad_state {};
    class Prelohi;
    class Lohi;
    // +lohi, -lohi
    Lohi operator+(const Lohi &);
    Lohi operator-(Lohi);
    // lohi + m, lohi - m, m - lohi
    Lohi operator+(Lohi, const Lohi &);
    Lohi operator+(Lohi, const mpz_class &);
    Lohi operator+(const mpz_class &, Lohi);
    Lohi operator-(Lohi, const Lohi &);
    Lohi operator-(Lohi, const mpz_class &);
    Lohi operator-(const mpz_class &, Lohi);
    // lohi * n, lohi / n
    Lohi operator*(Lohi, const Lohi &);
    Lohi operator*(Lohi, const mpz_class &);
    Lohi operator*(Lohi, long);
    Lohi operator*(const mpz_class &, Lohi);
    Lohi operator*(long, Lohi);
    Lohi operator/(Lohi, const mpz_class &);
    Lohi operator/(Lohi, long);
    Lohi operator/(const mpz_class &, Lohi);
    // lohi << k, lohi >> k
    Lohi operator<<(Lohi, mp_bitcnt_t);
    Lohi operator>>(Lohi, mp_bitcnt_t);
    // square(n)
    Lohi square(Lohi);
    mpz_class square(mpz_class);
    long square(long);
    // square(*n), negate(*n)
    Lohi &square(Lohi *);
    mpz_class &square(mpz_class *);
    long &square(long *);
    Lohi &negate(Lohi *);
    mpz_class &negate(mpz_class *);
    long &negate(long *);
    // delta, midpoint, deviation
    mpz_class delta(const Lohi &);
    mpz_class   mid(const Lohi &);
    mpz_class   dev(const Lohi &);
}

struct Interval::Prelohi {
private:
    mpz_class lo_;
    mpz_class hi_;
public:
    inline const mpz_class &lo() const noexcept { return lo_; }
    inline const mpz_class &hi() const noexcept { return hi_; }
    inline const mpz_class &ne() const noexcept { return lo_; }
    inline const mpz_class &fa() const noexcept { return hi_; }
    inline mpz_class &lo() noexcept { return lo_; }
    inline mpz_class &hi() noexcept { return hi_; }
    inline mpz_class &ne() noexcept { return lo_; } // near
    inline mpz_class &fa() noexcept { return hi_; } // far
    Prelohi() noexcept;
    explicit Prelohi(const mpz_class &) noexcept;
    Prelohi(const mpz_class &, const mpz_class &) noexcept;
    Prelohi &swap() noexcept;
};

class Interval::Lohi {
private:
    mpz_class lo_;
    mpz_class hi_;
public:
    inline const mpz_class &lo() const noexcept { return lo_; }
    inline const mpz_class &hi() const noexcept { return hi_; }
    Lohi() noexcept;
    explicit Lohi(const mpz_class &) noexcept;
    Lohi(const mpz_class &, const mpz_class &);
    Lohi(const Prelohi &);
    Lohi &operator=(const mpz_class &);
    // += -= *= /= <<= >>=
    Lohi &operator+=(const Lohi &);
    Lohi &operator+=(const mpz_class &);
    Lohi &operator-=(const Lohi &);
    Lohi &operator-=(const mpz_class &);
    Lohi &operator*=(const Lohi &);
    Lohi &operator*=(const mpz_class &);
    Lohi &operator*=(long);
    Lohi &operator/=(const mpz_class &);
    Lohi &operator/=(long);
    Lohi &operator<<=(mp_bitcnt_t);
    Lohi &operator>>=(mp_bitcnt_t);
    Lohi &square();
    Lohi &negate();
    Lohi &invert_against(const mpz_class &);
};
// The implementation correctly handles self-operations
// like lohi1 -= lohi1 and lohi2 *= lohi2.  However, these probably
// don't do what you want -- for, though they handle memory correctly,
// their *arithmetical logic* cannot assume that the two instances of
// the uncertain integer have the same value.  For example, if
// lohi1.lo() == 25 and lohi1.hi() == 28, then lohi1 -= lohi1 acts
// arithmetically as it were lohi1a -= lohi1b, whose lo() == -3 and
// hi() == 3.  Use lohi1 = 0 and lohi2.square(), rather.

#endif

