#include "roots.hh"
#include "divrem.hh"
#include "gcd.hh"
#include <cstdlib>
#include <utility>
#include <algorithm>

// This module was originally made to evaluate
//
//     x = (p + s*sqrt(n)) / q
//
// Though the module now does more, to understand the module it may
// help first to understand how the original problem was approached,
// as follows.
//
// Isolating the s*sqrt(n) in the above equation,
//
//     s*sqrt(n) = q*x - p
//
// Squaring, transferring all terms to the same side, defining f(x),
// and letting f(x) = 0,
//
//     f(x) = q*q*x*x - 2*q*p*x + p*p - s*s*n
//
// The desired root then is the value of x at which f(x) = 0 --
// provided that this value lies on the correct flank of the extremum
// of f(x).
//
// The extremum of f(x) occurs where df/dx = 0, so taking the last
// equation's derivative,
//
//     df/dx = 2*q*q*x - 2*q*p
//
// Reidentifying x as x_min, dividing by 2*q, setting the result
// to zero, and solving,
//
//     x_min = p/q
//
// How can one tell that it's x_min rather than x_max?  Answer:
// because d^2f/dx^2 = 2*q*q > 0.  At the extremum or in other
// words at the minimum,
//
//     f(x_min) = -s*s*n
//
// The principal weakness of the foregoing development is that, while
// isolating and squaring, it has lost information regarding the sign
// of s, which means that it does not know whether to find the lower or
// upper root of f(x).  Therefore, before following the above
// development, the program examines the sign of s relative to the
// sign of q to determine which root to find; and before
// doing *that*, it separately handles the special case of s == 0.
//
//

namespace {

    template <class T>
    mpz_class raise_to_nonnegative(long power, const T base) {
        mpz_class result {1};
        if (power < 0) throw Interval::Forbidden_negative {};
        for (; power; --power) result *= base;
        return result;
    };

    struct N_bools {
        bool was_n_negative;
        bool do_negate_f;
    };

    template <class T>
    inline N_bools construct_indicant_re_power_impl(
        long *const ptr_n, long *const ptr_m,
        const T &p,
        signed char *const ptr_flank
    )
    {
        Interval::reduce(ptr_n, ptr_m);
        if (*ptr_m <= 0) throw Interval::Impossibility {};
        const bool was_n_negative {*ptr_n < 0};
        if (!p && (!*ptr_n || was_n_negative))
          throw Interval::Division_by_zero {};
        *ptr_n = std::labs(*ptr_n);
        *ptr_flank = p ? (p < 0 ? -1 : 1) : 0;
        const bool is_n_odd(*ptr_n & 1);
        if (*ptr_m & 1) {
            if (!is_n_odd) *ptr_flank = std::abs(*ptr_flank);
        }
        else {
            if (is_n_odd) {
                if (*ptr_flank < 0) throw Interval::Forbidden_negative {};
            }
            else throw Interval::Impossibility {};
        }
        return N_bools
          { was_n_negative, (!was_n_negative && is_n_odd && p < 0) };
    }

}

Interval::Indicant::Indicant(
    const mp_bitcnt_t precision1, const long order1,
    const bool prefer_negative_branch1
):
  precision_ {precision1}, order_ {order1},
  prefer_negative_branch_ {prefer_negative_branch1}
{}

Interval::Indicant_re_general_quadratic::Indicant_re_general_quadratic
(
    const mp_bitcnt_t precision1,
    const long n, const long s, const long p, const long q,
    const bool prefer_negative_branch1
):
  Indicant(precision1, 2, prefer_negative_branch1),
  min_
(
    // Note:  the following throws Division_by_zero if q is zero.
    [&precision1, &p, &q]() -> Interval::Lohi {
        const Interval::Quotrem<mpz_class, long> quotrem
          {Interval::divrem(mpz_class(p) << precision1, q)};
        return Interval::Lohi {quotrem.quot, quotrem.quot + bool(quotrem.rem)};
    }()
),
  a2 {mpz_class(q)*q}, a1 {(mpz_class(-q)*p) << (precision1 + 1)},
  a0 {(mpz_class(p)*p - (mpz_class(s)*s)*n) << (precision1 << 1)},
  flank_
(
    [&s, &q, &prefer_negative_branch1]() -> signed char {
        const signed char flank1
          (s ? ((q >= 0 && s < 0) || (q < 0 && s >= 0) ? -1 : 1) : 0);
        return prefer_negative_branch1 ? -flank1 : flank1;
    }()
)
{ if (n < 0) throw Forbidden_negative {}; }

mpz_class Interval::Indicant_re_general_quadratic::operator()
  (const mpz_class &x) const
{ return (a2*x + a1)*x + a0; }

Interval::Lohi Interval::Indicant_re_general_quadratic::min() const
{ return min_; }

signed char Interval::Indicant_re_general_quadratic::flank() const
{ return flank_; }

Interval::Indicant_re_power::Indicant_re_power(
    const mp_bitcnt_t precision1,
    long n_, long m_, long p, long q,
    const bool prefer_negative_branch1
): Indicant(precision1, 0, prefer_negative_branch1)
{
    if (!q) throw Division_by_zero {};
    const bool was_q_negative {q < 0};
    if (was_q_negative) p = -p, q = -q;
    const N_bools n_bools
      {construct_indicant_re_power_impl(&n_, &m_, p, &flank_)};
    m      = m_;
    order_ = n_ + m_;
    if (!(m_ & 1) && prefer_negative_branch()) flank_ = -flank_;
    // If everything is positive, then f(x) = x^m * q^n - p^n.
    am =  raise_to_nonnegative(n_, (n_bools.was_n_negative ? p : q));
    a0 = -raise_to_nonnegative(n_, (n_bools.was_n_negative ? q : p))
      << (m_*precision());
    if (n_bools.do_negate_f) am = -am, a0 = -a0;
}

mpz_class Interval::Indicant_re_power::operator()
  (const mpz_class &x) const
{ return am*raise_to_nonnegative(m, x) + a0; }

Interval::Lohi Interval::Indicant_re_power::min() const
{ return Lohi {0}; }

signed char Interval::Indicant_re_power::flank() const
{ return flank_; }

Interval::Indicant_re_irrat_to_power::Indicant_re_irrat_to_power(
    const mp_bitcnt_t precision1,
    long n_, long m_, mpz_class irrat,
    const bool prefer_negative_branch1
): Indicant(precision1, 0, prefer_negative_branch1)
{
    const N_bools n_bools
      {construct_indicant_re_power_impl(&n_, &m_, irrat, &flank_)};
    was_n_negative = n_bools.was_n_negative;
    do_negate_f    = n_bools.do_negate_f;
    m              = m_;
    if (!(m_ & 1) && prefer_negative_branch()) flank_ = -flank_;
    if (n_bools.was_n_negative) {
        order_ = n_ + m_;
        // f(x) = x^m * irrat^n - 1
        coeff = raise_to_nonnegative(n_, irrat); //       am
        log2c = order_*precision();              // log2(-a0)
    }
    else if (n_ < m_) {
        order_ = m_;
        // f(x) = x^m - irrat^n
        log2c = 0;                               // log2( am)
        coeff = raise_to_nonnegative(n_, irrat)  //      -a0
          << ((m_-n_)*precision());
    }
    else {
        order_ = n_;
        // f(x) = x^m - irrat^n
        log2c = (n_-m_)*precision();             // log2( am)
        coeff = raise_to_nonnegative(n_, irrat); //      -a0
    }
}

mpz_class Interval::Indicant_re_irrat_to_power::operator()
  (const mpz_class &x) const
{
    // The writer would have used a ?: operator in place of the if-statement,
    // but for some apparently template-related reason the compiler objects.
    mpz_class result;
    if (was_n_negative)
      result = coeff*raise_to_nonnegative(m, x) - (mpz_class {1} << log2c);
    else
      result = (raise_to_nonnegative(m, x) << log2c) - coeff;
    return do_negate_f ? -result : result;
}

Interval::Lohi Interval::Indicant_re_irrat_to_power::min() const
{ return Lohi {0}; }

signed char Interval::Indicant_re_irrat_to_power::flank() const
{ return flank_; }

Interval::Lohi Interval::root(const Indicant &f)
{

    // This function might have been cleverer to use a Newton-Raphson
    // scheme but, since the number of iterations the code requires is
    // already of the order of the number of bits of accuracy the code
    // seeks, the present, simpler scheme suffices.  (Cleverness comes
    // at a cost.  Code is maintainable only to the extent to which a
    // maintainer that understands the code's cleverness, or that is
    // willing to devote the time necessary to study the code's
    // cleverness, is available.  There are times at which such
    // cleverness is better avoided and this is probably one of
    // those times.)

    const Lohi x_min {f.min()};
    const signed char flank {f.flank()};

    // Dispose of the nonradical degenerate case.
    if (!flank) return x_min;

    Prelohi x;
    {

        // Initialize x1 to the minimum.
        mpz_class x1 {flank < 0 ? x_min.lo() : x_min.hi()};

        // Treat the presumably rare case in which the root happens
        // to lie immediately next to the minimum.
        {
            const mpz_class f1 {f(x1)};
            if (f1 >= 0) {
                x.ne() = x1 - (f1 ? flank : 0);
                x.fa() = x1;
                //
                // A certain, celebrated, stubborn, influential
                // professor and early computer scientist (to name him
                // here would be captious), though intelligent and
                // original, was unfortunately a fount of bad
                // programming advice -- not only regarding the goto
                // statement but regarding procedural exits generally.
                // Whatever his contributions in the field of
                // algorithms may have been, these are outweighed by
                // the persistent annoyance experienced programmers
                // suffer in correction by sophomores the professor's
                // teachings have misled.  Rules confirmed by time and
                // experience are good; rules contradicted by time and
                // experience are bad; rules whose profitable
                // contradiction the rulemaker had years to observe,
                // but which the rulemaker nevertheless refused to
                // withdraw, are worse.  The professor in question was
                // entitled to his opinion but the present writer would
                // not recommend adherence to the professor's rules.
                //
                goto has_found;
                //
                // Some reader will inevitably object, "But you could
                // have done that without the goto!"  But that is just
                // it:  there are many programming elements one *could*
                // do without.  That one could do without an element is
                // hardly sufficient reason to do without it, nor is
                // the proscription of goto even supported by an
                // aesthetic appeal to minimalism, since goto is about
                // as minimal as one can get.  The aforementioned
                // professor, of course, gave a reason for his
                // proscription.  The reason was even plausible, only
                // wrong.  More importantly, the reason has been
                // extensively contradicted by millions of lines of
                // subsequent good code.
                //
                // The goto statement since 1990 or so is best
                // understood as a simple programming element whose
                // practical occasion for use happens to be fairly
                // rare -- arising maybe once per 5000 lines of code or
                // so.  It's not that one should avoid it, but rather
                // that the circumstance in which one would even think
                // to use it does not occur too often.  It's a bit
                // like, say, C++'s pointer-to-member feature -- fine,
                // good to use when and as needed, but not frequently
                // met.  There was a context during the 1970s
                // and 1980s, having to do with the transition from
                // primitive early programming languages to better ones
                // like C++'s ancestor C, in which the professor's
                // advice seemed to make some sense.  The writer is
                // just old enough to remember those days and is
                // skeptical that the advice was good even then; but
                // the professor, now deceased, remained professionally
                // active throughout the 1990s and into the 21st
                // century, long after his advice's uselessness had
                // grown broadly apparent.  He should have
                // withdrawn it.
            }
        }

        // Leap longer and longer until the root has been leapt past.
        mpz_class delta {flank};
        while (true) {
            const mpz_class x1_new {x1 + delta};
            const mpz_class f1 {f(x1_new)};
            if (f1 >= 0) {
                if (f1 == 0) {
                    x.ne() = x1_new;
                    x.fa() = x1_new;
                    goto has_found;
                }
                break;
            }
            x1 = x1_new;
            delta <<= 1;
        }

        // The root now lies between x1 and x1 + delta, exclusive.
        // Zero in on it.
        while (delta > 1 || delta < -1) {
            delta >>= 1;
            const mpz_class x1_new {x1 + delta};
            const mpz_class f1 {f(x1_new)};
            if (f1 == 0) {
                x.ne() = x1_new;
                x.fa() = x1_new;
                goto has_found;
            }
            else if (f1 < 0) {
                x1 = x1_new;
            }
        }
        x.ne() = x1;
        x.fa() = x1 + flank;

    }

has_found:
    if (flank < 0) x.swap();
    return x;

}

Interval::Lohi Interval::root(
    const mp_bitcnt_t precision, long n, long m, const Lohi &x,
    const bool prefer_negative_branch
)
{
    reduce(&n, &m);
    // The mathematical expression 0^0 without context is ambiguous,
    // so Indicant's subclasses reject it.  However, in the present
    // context, the expression means (0)^[0] == 1, in which
    // the (0) is inexact and the [0] is exact.  The next line
    // therefore bypasses Indicant in the 0^0 case.
    if (!n) return Lohi {mpz_class {1} << precision};
    const bool does_x_straddle {x.lo() < 0 && x.hi() > 0};
    const bool is_x_negative {x.lo() < 0 && x.hi() <= 0};
    if (does_x_straddle && is_x_negative) throw Impossibility {};
    const bool is_n_negative {n < 0};
    const bool is_n_odd(n & 1);
    const auto root_of {
        [&precision, &x, &n, &m, &prefer_negative_branch]
        (const mpz_class &(Lohi::*const ptr)() const) -> Lohi
        {
            return root(
                Indicant_re_irrat_to_power
                (precision, n, m, (x.*ptr)(), prefer_negative_branch)
            );
        }
    };
    const auto straight_result {
        [&root_of] () -> Lohi
        { return Lohi {root_of(&Lohi::lo).lo(), root_of(&Lohi::hi).hi()}; }
    };
    const auto reversed_result {
        [&root_of] () -> Lohi
        { return Lohi {root_of(&Lohi::hi).lo(), root_of(&Lohi::lo).hi()}; }
    };
    const auto nonnegative_result {
        [&root_of] () -> Lohi
        {
            return Lohi
            {0, std::max(root_of(&Lohi::lo).hi(), root_of(&Lohi::hi).hi())};
        }
    };
    //    n/m      -x       ?x       +x
    // +odd/odd straight straight straight
    // -odd/odd reversed forbiddn reversed
    // +evn/odd reversed nonnegtv straight
    // -evn/odd straight forbiddn reversed
    // +odd/evn forbiddn forbiddn straight branching
    // -odd/evn forbiddn forbiddn reversed branching
    if (m & 1) {
        if (is_n_odd) {
            //    n/m      -x       ?x       +x
            // +odd/odd straight straight straight
            // -odd/odd reversed forbiddn reversed
            if (is_n_negative) {
                if (does_x_straddle) throw Division_by_zero {};
                return reversed_result();
            }
            else {
                return straight_result();
            }
        }
        else {
            //    n/m      -x       ?x       +x
            // +evn/odd reversed nonnegtv straight
            // -evn/odd straight forbiddn reversed
            if (does_x_straddle) {
                if (is_n_negative) throw Forbidden_negative {};
                return nonnegative_result();
            }
            else {
                return
                ( is_x_negative &&  is_n_negative) ||
                (!is_x_negative && !is_n_negative)
                ? straight_result() : reversed_result();
            }
        }
    }
    else {
        if (is_n_odd) {
            //    n/m      -x       ?x       +x
            // +odd/evn forbiddn forbiddn straight branching
            // -odd/evn forbiddn forbiddn reversed branching
            if (is_x_negative || does_x_straddle) throw Forbidden_negative {};
            return
              ( is_n_negative &&  prefer_negative_branch) ||
              (!is_n_negative && !prefer_negative_branch)
              ? straight_result() : reversed_result();
        }
        else throw Impossibility {};
    }
    throw Impossibility {};
    return Lohi {};
}

